
// High level message dispatch to native code. If we are running in WebKit2, we'll have the expected
// window.webkit.messageHandler.rsHTMLEditor object. Otherwise we assume we'll have legacy WebView
// script handler attached.
function rsPostMessage(message) {
	var webkit = window.webkit;
	if (!message.parameters) {
		message.parameters = [];
	}
	if (webkit && webkit.messageHandlers.rsHTMLEditor) {
		webkit.messageHandlers.rsHTMLEditor.postMessage(message);
	}
	else {
		window.rsHTMLEditor.handleLegacyWebViewMessage_(JSON.stringify(message));
	}
}

function gotJavaScriptError(errorEvent) {
	console.log("Error: " + errorEvent.message);
	message = new Object();
	message['name'] = "error";
	message['parameters'] = [errorEvent.message];
	rsPostMessage(message);
}

function rsDebugLog(debugText) {
	rsPostMessage({'name': 'log', 'parameters': [debugText]});
}

function rsSelectedNode() {
	var selection = rsEditor.selection.getSel();
	if (selection.rangeCount === 0) { return null; }
	var selectionRange = selection.getRangeAt(0);
	return selectionRange.commonAncestorContainer;
}

function rsSelectedElement() {
	var selectionNode = rsSelectedNode();
	if (!selectionNode) {
		return null;
	}
	if (selectionNode.nodeType == Node.ELEMENT_NODE) {
		return selectionNode;
	}
	else {
		return selectionNode.parentElement;
	}
}

function rsNodePathForSelection() {
	var selectionNode = rsSelectedNode();
	var nodePath = [];
	var nextNode = selectionNode;
	while (nextNode) {
		nodePath.unshift(nextNode.nodeName)
		nextNode = nextNode.parentNode;
	}
	return nodePath;
}

function rsCommandStatesForSelection() {
	var selectedElement = rsSelectedElement();
	return {
		"bold": tinymce.editors.RSHTMLContent.queryCommandState("Bold"),
		"italic": tinymce.editors.RSHTMLContent.queryCommandState("Italic"),
		"underline": tinymce.editors.RSHTMLContent.queryCommandState("Underline"),
		"blockQuote": tinymce.editors.RSHTMLContent.formatter.match('blockquote'),
		"paragraph": tinymce.editors.RSHTMLContent.formatter.match('p'),
		"preformatted": tinymce.editors.RSHTMLContent.formatter.match('pre'),
		"strikethrough": tinymce.editors.RSHTMLContent.formatter.match('strikethrough'),
		"h1": tinymce.editors.RSHTMLContent.formatter.match('h1'),
		"h2": tinymce.editors.RSHTMLContent.formatter.match('h2'),
		"h3": tinymce.editors.RSHTMLContent.formatter.match('h3'),
		"h4": tinymce.editors.RSHTMLContent.formatter.match('h4'),
		"h5": tinymce.editors.RSHTMLContent.formatter.match('h5'),
		"h6": tinymce.editors.RSHTMLContent.formatter.match('h6'),
		"alignLeft": tinymce.editors.RSHTMLContent.queryCommandState("JustifyLeft"),
		"alignCenter": tinymce.editors.RSHTMLContent.queryCommandState("JustifyCenter"),
		"alignRight": tinymce.editors.RSHTMLContent.queryCommandState("JustifyRight"),
		"alignJustified": tinymce.editors.RSHTMLContent.queryCommandState("JustifyFull"),
		"editLink": tinymce.editors.RSHTMLContent.queryCommandState("EditLink"),
		"undo": rsEditor.undoManager.hasUndo(),
		"redo": rsEditor.undoManager.hasRedo()
	};
}

function rsSelectionStateJSON() {
	// So that the client can reckon the selection rect (expressed in terms of the whole document)
	// relative to the scroll view's content area, we include both the bounding rect and the
	// corresponding
	var documentRect = rsEditor.getDoc().documentElement.getClientRects()[0];
	documentRect.x = 0;
	documentRect.y = 0;
	var clientObject = {"editorBounds": documentRect,
						"selectionFrame": window.rsSelectionBoundingRect(),
						"selectionMarkup": rsEditor.selection.getContent(),
						"selectionText": rsEditor.selection.getSel().toString(),
						"selectionNodePath": rsNodePathForSelection(),
						"selectionLinkHREF": rsGetSelectedAnchorNodeHREFString(),
						"commandStates": rsCommandStatesForSelection()
						};

	return JSON.stringify(clientObject);
}

// event may be null - for example we fabricate a selection changed notification
// after toggling typewriter mode to account for the changed rect. We might want to
// rename this to something more like "selection rect changed".
function rsSendSelectionChangedNotification(event) {
	if (rsDisableNotifications === true) {
		return;
	}

	var selectionStateJSON = rsSelectionStateJSON();
	if (selectionStateJSON !== window.rsLastSelectionChangedInfo) {
		var message = Object();
		message['name'] = "selectionChanged";
		message['parameters'] = [selectionStateJSON];
		rsPostMessage(message);
	}
}

// Used for both content change notifications and as the return object for methods
// that have mutation side-effects. In the case of a mutation side-effect the notification
// is not sent, to avoid redundant notifications. The caller from the native side is expected
// to treat the resulting object the same as if it was notified of content change.
function rsContentSnapshot() {
	// We combine the selection state and the current source string
	var selectionStateJSON = rsSelectionStateJSON();
	var sourceString = rsEditor.getContent({source_view : true});
	return [selectionStateJSON, sourceString]
}

function rsSendContentChangedNotification() {
	if (rsDisableNotifications === true) {
		return;
	}

	var message = Object();
	message['name'] = "contentChanged";
	message['parameters'] = rsContentSnapshot();
	rsPostMessage(message);
}

// To facilitate the editor handling pastes differently when they come from within the app vs.
// from without, we tag copy/cut with extra custom pasteboard data to "tag" it as coming from us.
function rsPasteboardWritingEventHandler(event) {
	// WKWebView doesn't allow web content to set arbitrary pasteboard values, so call out to the native code
	// SMELL(DCJ): Maybe the whole copy process should be driven from here?
	rsPostMessage({'name': 'pasteboardEvent', 'parameters': []});
}

// Highest level of interception happens at the app level where the whole paste:
// action is delegated to the RSHTMLEditView delegate and it can take the data
// right off the pasteboard itself, for example to facilitate a media upload. When
// a paste reaches here it can be specialized depending on what kind of paste it is,
// and what is on the clipboard.
function rsPasteboardReadingEventHandler(event) {
	// If the pasteboard includes x-tinymce/html data, it means the copy originated
	// from our editor, so we allow the usual paste to happen with full fidelity. If
	// not, we default to pasting as plain text.
	if (!window.rsExpectingRichTextPaste) {
		// Defer to TinyMCE if there is TinyMCE-originated HTML, or if there is a URL on the pasteboard,
		// since pasting a URL causes linking the selection to that URL.
		function isURLString(aString) {
			// Could try to initialize a 'new URL(...)' but I'm not sure I want to be as lenient as that
			// for the purposes of deciding what to reasonably link to.
			return aString.startsWith("https://") || aString.startsWith("http://");
		}
		function clipboardHasURL(clipboardData) {
			return clipboardData.types.includes("text/uri-list") || isURLString(String(event.clipboardData.getData("text/plain")));
		}
		if (!event.clipboardData.types.includes("x-tinymce/html") && !clipboardHasURL(event.clipboardData)) {
			// Also require actually having plain text to insert
			var plainText = event.clipboardData.getData("text/plain");
			if (plainText) {
				rsEditorDoc.execCommand("insertText", false, plainText);
				event.preventDefault();
				event.stopImmediatePropagation();
			}
		}
	}
	window.rsExpectingRichTextPaste = false;
}

function rsDragEventHandler(event) {
	// rsPostMessage({'name': 'log', 'parameters': ['Got drag']});

	// Keep track of the corresponding drop location for the caret insertion point.
	// Note this method is non-standard but as of 2022 Safari implements it while still
	// not yet implementing the standards/experimental Document.caretPositionFromPoint
	let nextInsertionPoint = rsEditorDoc.caretRangeFromPoint(event.clientX, event.clientY);
	if (nextInsertionPoint) {
		window.rsDroppedItemInsertionPoint = nextInsertionPoint;
//		rsDebugLog("window drop point now " + describeRange(nextInsertionPoint));
	}
}

function rsPreserveLastSelectionRange() {
	// Remember the last selection range so we can restore it whenever we regain focus
	const selection = rsEditor.getWin().getSelection();

	let nextRange = selection.rangeCount === 0 ? null : selection.getRangeAt(0);
	if (nextRange) {
		saveLastSelectionRange(nextRange);
	}
}

//function describeRange(range) {
//	if (!range) { return "NULL"; }
//	return range.startOffset.toString();
//}
//
//function describeSelection(sel) {
//	if (!sel) { return "NULL"; }
//	return sel.focusNode.toString() + " / " + sel.focusOffset.toString();
//}

function rsRestoreLastSelectionRange() {
	if (window.rsLastSelectionRange) {
		var nextRange = window.rsLastSelectionRange.cloneRange();
		var sel = rsEditor.getWin().getSelection();
		sel.removeAllRanges();
		sel.addRange(nextRange);

		// Trying to alleviate some weird issues where the rsLastSelectionRange seems to turn to
		// garbage, maybe because of edits to the document? Trying to be a little more aggressive
		// about only ever restoring a range once, and relying on a more proximal save of existing range
		// for any other restore that might be necessary.
		window.rsLastSelectionRange = null;
	}
}

// Bottleneck for saving the range so we can sanitize behavior
function saveLastSelectionRange(range) {
	let savedRange = null;
	if (range) {
		savedRange = range.cloneRange();
	}
	window.rsLastSelectionRange = savedRange;
	//rsDebugLog("Preserved range " + describeRange(savedRange));
}

function myCustomSetupContent(setupEvent)
{
	rsSetEditorColors(gRSInitialForegroundColor, gRSInitialBackgroundColor);

	// Set up a bunch of global variables for the benefit of easy access. Would probably
	// be better encapsulated in a single top-level object (maybe window.rsEditor should
	// contain the TinyMCE editor and everything else we track).
	window.rsEditor = setupEvent.target;
	window.rsEditorDoc = rsEditor.getDoc();
	window.rsTypewriterModeManager = {"enabled": false};
	window.rsDisableNotifications = false;
	window.rsLastSelectionChangedInfo = "";
	window.rsDroppedItemInsertionPoint = null;
	window.rsExpectingRichTextPaste = false;
	saveLastSelectionRange(null);

	// The rich editor has a separate 'Window' that we need to also install the event listener on
	rsEditor.getDoc().defaultView.addEventListener("error", gotJavaScriptError, false);

	// Currently we don't actually override these actions in a meaningful way because the
	// Cut and Copy commands are handled by WebKit and thus we notice and append from
	// webView:didWriteSelectionToPasteboard: in RSHTMLEditView+WebEditingDelegate.m
	// Leaving these hooks in place as a reminder and stopgap to handle them correctly
	// if/when TinyMCE again asserts ownership over handling of these commands.
	rsEditor.getDoc().defaultView.addEventListener("cut", rsPasteboardWritingEventHandler, false);
	rsEditor.getDoc().defaultView.addEventListener("copy", rsPasteboardWritingEventHandler, false);
	rsEditor.getDoc().defaultView.addEventListener("paste", rsPasteboardReadingEventHandler, true);
	rsEditor.getDoc().defaultView.addEventListener("drag", rsDragEventHandler, true);
	rsEditor.getDoc().defaultView.addEventListener("dragover", rsDragEventHandler, true);

	if (gRSInitializeInDarkMode) {
	  switchToDarkMode();
	}

	// Ensure the rich editor is selected
	selectRichEditorIfNeeded();

	// Install selection change notifier
	rsEditorDoc.addEventListener("selectionchange", function(event) {
		rsSendSelectionChangedNotification(event);
	});

	rsEditorDoc.body.addEventListener("focus", function(event) {
		rsRestoreLastSelectionRange();
	});

	rsEditorDoc.body.addEventListener("mousedown", function(event) {
		// Clear the preserved selection range so that the selection
		// will change to match the click location, instead of
		// being reset in focus above.
		saveLastSelectionRange(null);
	});

	rsEditorDoc.body.addEventListener("mouseup", function(event) {
		// Clear the preserved selection range so that the selection
		// will change to match the click location, instead of
		// being reset in focus above.
		saveLastSelectionRange(null);
	});

	// Let RSHTMLEditView know that we're done setting up
	var didFinishMessage = Object();
	didFinishMessage['name'] = "finishedLoading";
	rsPostMessage(didFinishMessage);

	// Install click listeners on the main document to avoid shifting focus out of the
	// editor component and to the containing web view. This is not ordinarily an issue
	// if the editor takes up all the space of the web view, but just in case...
	function interceptClick(event) {
		event.preventDefault();
		event.stopImmediatePropagation();
	}
	document.addEventListener("mousedown", interceptClick, true);
	document.addEventListener("click", interceptClick, true);

	window.addEventListener("resize", function(event) {
		rsUpdateScrollPositionForTypewriterMode();
	 });

	// Handle tab keys specialy
	function tabKeyFilter(event) {
		if ((event.keyCode === 9) &&
			(!event.metaKey) &&
			(!event.altKey) &&
			(!event.ctrlKey))
		{
			var isKeyDown = (event.type == "keydown");

			// Don't filter when inside an element where tabs are meaningful
			var isInList = false;
			var isInPre = false;
			var selectionElement = selectionCommonAncestorElement();
			if (selectionElement !== null) {
				isInList = Boolean(selectionElement.closest("li"));
				isInPre = Boolean(selectionElement.closest("pre"));
			}
			if (isInList || isInPre) {
				// Outdenting/Indenting happens OK by default, but in preformatted text
				// we seem to need to impose the literal tab character.
				if (isInPre && isKeyDown) {
					rsEditor.insertContent("\t");
				}
				return;
			}

			// Only convey the keydown event to the client, but consume the key up counterpart in any case
			if (isKeyDown) {
				var isShiftParam = event.shiftKey ? "true" : "false";
				rsPostMessage({'name': 'unconsumedTabKey', 'parameters': [isShiftParam]})
			}
			event.preventDefault();
		}
	}
	rsEditorDoc.body.addEventListener('keydown', tabKeyFilter, true);
	rsEditorDoc.body.addEventListener('keyup', tabKeyFilter, true);

	// Observe mutations to the editor content
	rsObserveDOMChanges(rsEditor.getBody(), function() {
		self.rsSendContentChangedNotification();
	});
}

// From https://stackoverflow.com/questions/3219758/detect-changes-in-the-dom
var rsObserveDOMChanges = (function(){
  var MutationObserver = window.MutationObserver

  return function( obj, callback ){
	if( !obj || obj.nodeType !== 1 ) return;

	if( MutationObserver ){
	  // define a new observer
	  var mutationObserver = new MutationObserver(callback)

	  // Aggressively monitor all changes recursively including text content and style attributes
	  mutationObserver.observe( obj, { childList:true, subtree:true, characterData:true, attributes: true, attributesFilter: ["style"] })
	  return mutationObserver
	}
  }
})()

function selectRichEditor() {
	if (window.rsEditor) {
		rsEditor.selection.select(rsEditor.getBody(), 1);
		rsEditor.selection.collapse(1);
	}
}

function selectRichEditorIfNeeded() {
	if (window.rsEditor) {
		var activeElement = document.activeElement;
		if ((activeElement === null) || (rsEditor.iframeElement.contains(activeElement) == false)) {
			selectRichEditor();
		}
	}
}

// Invoked by drag/drop infrastructure to ask us to change the selection to the last recorded
// location of the drag insertion caret.
function rsMoveSelectionToDragCaretPosition(selectByWord) {
	var caretPoint = window.rsDroppedItemInsertionPoint;
	if (caretPoint) {
		rsEditor.selection.setRng(caretPoint);
//		rsDebugLog("moved selection range to " + describeRange(caretPoint));
		window.rsPreserveLastSelectionRange();
	}
}

function replaceSelectionWithMarkupString(markupString, selectingReplacement = true) {
	var ed = tinymce.editors.RSHTMLContent;

	// Because we may have been unfocused to, for example, open the Media Manager in MarsEdit,
	// we need to be careful to restore the previous selection so that the replacement occurs
	// at the expected location.
	rsRestoreLastSelectionRange();

	// Because executing a command will actually create an undo level, we wrap the whole thing in a
	// transaction to ensure the undo manager treats the whole thing as an atomic change and ignores
	// undo registrations made within e.g. the execCommand that sets the selection
	ed.undoManager.transact(function () {
		var selection = ed.selection.getSel();

		// The following code will cause the selection to become literally replaced, i.e. the new markup
		// becomes the selection.  This is awkward especially for e.g. images that have been inserted,
		// but could be valid if for instance we were replacing some selection with a new text style.
		//
		// The compromise we take is if there was a non-caret selection, then we replace it and leave
		// it, otherwise we move the insertion point to just past the inserted item.
		var hadSelection = selection.isCollapsed === false;

		// TinyMCE's default behavior is to collapse after inserting content, so if
		// we want to keep the selection preserved, we need to save and restore via
		// the TinyMCE bookmarking mechanism.
		var bookmark = null;
		if (hadSelection && selectingReplacement) {
			bookmark = tinymce.editors.RSHTMLContent.selection.getBookmark();
		}

		// Lean on TinyMCE's insert content function to do the heavy lifting
		ed.execCommand('mceInsertContent', false, markupString);

		// Collapse selection unless we were meant to keep it selected
		if (hadSelection && !selectingReplacement) {
			tinymce.editors.RSHTMLContent.selection.collapse()
		}

		if (bookmark) {
			tinymce.editors.RSHTMLContent.selection.moveToBookmark(bookmark);
		}
	});
	return tinymce.editors.RSHTMLContent.getContent({source_view : true})
}

function selectionCommonAncestorElement() {
	var commonElement = null;

	var ed = tinymce.editors.RSHTMLContent;
	var selection = ed.selection.getSel();
	if (selection.rangeCount < 1) {
		return null;
	}
	var selectionRange = selection.getRangeAt(0);
	var selectionNode = selectionRange.commonAncestorContainer;
	if (selectionNode !== null) {
		commonElement = selectionNode.nodeType == 1 ? selectionNode : selectionNode.parentElement;
	}
	return commonElement;
}

function selectedAnchorNode() {
	var anchorNode = null;
	var selectionElement = selectionCommonAncestorElement();
	if (selectionElement !== null) {
		anchorNode = selectionElement.closest("a");
	}
	return anchorNode;
}

function rsGetSelectedAnchorNodeHREFString() {
	var foundHREF = null;

	var anchorNode = self.selectedAnchorNode();
	if (anchorNode !== null) {
		// This fragility reflects my uncertainty about whether TinyMCE stores the value
		// in the proper href attribute or a sneaky "data-mce-href" attribute. Try plain
		// href first.
		foundHREF = anchorNode.getAttribute("href");
		if (foundHREF === null) {
			foundHREF = anchorNode.getAttribute("data-mce-href");
		}
	}
	if (foundHREF === null) {
		return "";
	}
	else {
		return foundHREF;
	}
}

function rsLinkSelectionToHREFString(hrefString, deriveMissingLinkText = false) {
	rsDisableNotifications = true;

	var ed = tinymce.editors.RSHTMLContent;
	var selection = ed.selection.getSel();
	if (selection.rangeCount < 1) { return; }
	var selectionRange = selection.getRangeAt(0);

	// Are we already in an anchor node?
	var inAnchorNode = self.selectedAnchorNode() !== null;

	ed.undoManager.transact(function () {
		// If there is no selection, and we are not in a link, we insert and select the text of the href to make it a self-describing link
		// (NOTE: This is less than ideal but it does match what Mail.app does so some users will find it familiar).
		var isCollapsed = selection.isCollapsed;
		if (deriveMissingLinkText && isCollapsed && (inAnchorNode === false) && (hrefString.length > 0)) {
			var textNode = document.createTextNode(hrefString);
			selectionRange.insertNode(textNode);
			selection.extend(textNode);
		}

		// If we're not in an anchor node, don't bother setting link with empty text. That would just cause MCE to insert some bogus junk.
		if (inAnchorNode || (hrefString.length > 0)) {
			ed.execCommand('mceInsertLink', '', hrefString);
		}
	});

	rsDisableNotifications = false;

	return rsContentSnapshot();
}

function rsSetEditorBaseURL(urlString) {
	// We only ever expect to find at most 1 existing <base> node - the one we may have put in there previously.
	// Just remove it so we can add a new one.
	var ed = tinymce.editors.RSHTMLContent;
	var headElement = ed.getDoc().head;
	var baseNodeElements = headElement.getElementsByTagName('base');
	if (baseNodeElements.length > 0) {
		var baseElement = baseNodeElements.item(0);
		if (baseElement !== null) {
			baseElement.parentElement.removeChild(baseElement);
		}
	}

	headElement.insertAdjacentHTML('beforeend', '<base href="' + urlString + '" />');

	// WebKit doesn't do as good a job as it should in updating all the contents after the base changes,
	// so trigger an update by manually resetting the content to what it already is
	ed.setContent(ed.getContent({source_view : true}));

	return "";
}


function rsUpdateAttributes(attributeName, oldValue, newValue) {
	// Find any element whose named attribute matches the oldValue
	let xpath = "//*[@" + attributeName + "='" + oldValue + "']";
	let result = rsEditorDoc.evaluate(xpath, rsEditorDoc.body, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
	for (let i = 0; i < result.snapshotLength; i++) {
		let node = result.snapshotItem(i);
		node.setAttribute(attributeName, newValue);
	}
}

function rsUpdateLinkAttributes(oldURLString, newURLString) {
	var ed = tinymce.editors.RSHTMLContent;
	ed.undoManager.transact(function () {
		self.rsUpdateAttributes("href", oldURLString, newURLString);
		self.rsUpdateAttributes("src", oldURLString, newURLString);

		/// TinyMCE stashes redundant copy of the value in a custom attribute, and respects that when serializing
		self.rsUpdateAttributes("data-mce-src", oldURLString, newURLString);
		self.rsUpdateAttributes("data-mce-href", oldURLString, newURLString);
	});
}

function rsRemoveElementsWithAttributeValue(attributeName, attributeValue) {
	let xpath = "//*[@" + attributeName + "='" + attributeValue + "']";
	let result = rsEditorDoc.evaluate(xpath, rsEditorDoc.body, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
	for (let i = 0; i < result.snapshotLength; i++) {
		let node = result.snapshotItem(i);
		let parentElement = node.parentElement;
		if (parentElement) {
			parentElement.removeChild(node);
		}
	}
}

function rsRemoveElementsReferencingAttachment(urlString) {
	var ed = tinymce.editors.RSHTMLContent;
	ed.undoManager.transact(function () {
		self.rsRemoveElementsWithAttributeValue("href", urlString)
		self.rsRemoveElementsWithAttributeValue("src", urlString)
		self.rsRemoveElementsWithAttributeValue("data-mce-src", urlString)
		self.rsRemoveElementsWithAttributeValue("data-mce-href", urlString)
	});
}

function rsSetDefaultFont(familyName, pointSize) {
	rsEditor.getBody().style.fontFamily = familyName;
	rsEditor.getBody().style.fontSize = pointSize + "px";
}

function rsSetEditorColors(fgColorHex, bgColorHex) {
	var ed = tinymce.editors.RSHTMLContent;
	var bodyStyle = ed.getDoc().body.style;
	bodyStyle.color = fgColorHex;
	bodyStyle.backgroundColor = bgColorHex;

	// Setting the background color on the frame helps to ensure the right color
	// for the scroller in "always show scroll bar" mode, and sets the appropriate
	// overscroll color to match the background of the body.
	var frameStyle = ed.getWin().frameElement.style;
	frameStyle.backgroundColor = bgColorHex;
	return "";
}

// Wrapper on TinyMCE setContent that returns the properly massaged source string
function rsSetContent(sourceString) {
	rsEditor.setContent(sourceString);
	return rsEditor.getContent({source_view : true});
}

// DOM Utilities

// Bottleneck for customizing execCommand behavior. Eventually all our execCommand should come through here
// instead of trying to guess whether to direct the command at WebKit, the TinyMCE editor, native window, etc.
function rsExecCommand(commandName, ...params) {
	// This is a placeholder shim for now that just passes all commands through to the editor
	return rsEditor.execCommand(commandName, ...params);
}

// Convenience method first queries the DOM for the current value for the command, and only
// invokes the command with the new value if it will cause a change
function rsExecChangeValueCommandIfChanged(command, stringValue) {
	var currentValue = tinymce.editors.RSHTMLContent.queryCommandValue(command);
	if (currentValue !== stringValue) {
		return rsExecCommand(command, false, stringValue);
	}
	return false;
}

function rsTraverseNodes(baseNode, traverseBackward = false, skipBaseNode = true, callback) {
	var shouldStop = false;
	var ignoreChildren = false;
	var nextNode = baseNode;

	if (skipBaseNode) {
		nextNode = traverseBackward ? baseNode.previousSibling : baseNode.nextSibling;
		if (nextNode === null) {
			nextNode = baseNode.parentNode;
			ignoreChildren = true;
		}
	}

	while (nextNode !== null) {
		// If the node has no children or we're ignoring them, visit the node now
		if (ignoreChildren || nextNode.childNodes.length === 0) {
			if (callback(nextNode)) {
				break;
			}
		}
		else {
			// Dig deeply through extent of first or last children
			while (nextNode.childNodes.length > 0) {
				nextNode = traverseBackward ? nextNode.lastChild : nextNode.firstChild;
			}

			// Visit the leaf node
			if (callback(nextNode)) {
				break;
			}
		}

		// Move on to siblings, if we have them
		ignoreChildren = false;
		var siblingNode = traverseBackward ? nextNode.previousSibling : nextNode.nextSibling;
		if (siblingNode !== null) {
			nextNode = siblingNode;
		}
		else {
			// If no siblings, move up to the parent, noting that we should ignore its children
			// as we've already visited them.
			nextNode = nextNode.parentNode;
			ignoreChildren = true;
		}
	}
}

function rsAdjacentDepthFirstTextNode(baseNode, traverseBackward = false) {
	var foundNode = null
	rsTraverseNodes(baseNode, traverseBackward = traverseBackward, skipBaseNode = true, function (visitedNode) {
		if ((visitedNode.nodeType == 3) && (visitedNode != baseNode)) {
			foundNode = visitedNode;
			return true
		}
		else {
			return false;
		}
	});
	return foundNode;
}

// Page Break utilities

function rsPageBreakNode() {
	var ed = tinymce.editors.RSHTMLContent;
	return ed.getDoc().getElementById('com_red-sweater_htmleditor_pagebreak')
}


function rsInsertPageBreakIfNeeded(pageBreakMarkup) {
	if (rsPageBreakNode() === null) {
		// Append a blank paragraph to ensure we will have something to select/edit after the uneditable div
		var markupToInsert = pageBreakMarkup + "<p></p>";

		replaceSelectionWithMarkupString(markupToInsert);
	}
	
	return tinymce.editors.RSHTMLContent.getContent({source_view : true});
}

function rsRemovePageBreakIfNeeded() {
	var pageBreakNode = rsPageBreakNode();
	if (pageBreakNode !== null) {
		pageBreakNode.parentElement.removeChild(pageBreakNode);
	}

	return tinymce.editors.RSHTMLContent.getContent({source_view : true});
}

// Move to either just before or just after the page break, if present
function rsMoveToPageBreakNode(positionAfter = false) {
	var pageBreakNode = rsPageBreakNode();
	if (pageBreakNode !== null) {
		var foundTextNode = null;
		if (positionAfter === false) {
			foundTextNode = rsAdjacentDepthFirstTextNode(pageBreakNode, true)
		}
		else {
			foundTextNode = rsAdjacentDepthFirstTextNode(pageBreakNode, false)
		}
		if (foundTextNode != null) {
			let doc = ed.getDoc();
			let sel = doc.getSelection();
			var newSelectionRange = doc.createRange();
			newSelectionRange.selectNode(foundTextNode);
			sel.removeAllRanges();
			sel.addRange(newSelectionRange);
			if (positionAfter) {
				sel.collapseToStart();
			}
			else {
				sel.collapseToEnd();
			}
		}
	}
	// Accommodate the limited capability of the invocation method to handle return types that aren't strings ... just gets ignored.
	return "";
}

function rsUpdateScrollPositionForTypewriterMode() {
	var tmManager = window.rsTypewriterModeManager;
	if (tmManager && tmManager.enabled) {
		var rect = rsSelectionBoundingRect();

		// Computation is not quite right but this fudge factor makes it appear more centered.
		// Maybe this is approximately the line height?
		var kFudgeFactor = 10;

		// Scroll the selection bounding rect so that it is centered vertically
		var selectionCenterOffset = rect.bottom - (rect.height / 2.0) + kFudgeFactor;
		var centerScrollPosition = selectionCenterOffset - (window.innerHeight / 2.0) + kFudgeFactor;
		rsScrollToVerticalOffset(centerScrollPosition);
	}
}
function rsSetTypewriterModeEnabled(enabled) {
	var newEnabled = Boolean(enabled);
	var tmManager = window.rsTypewriterModeManager;
	if (tmManager.enabled !== newEnabled) {
		tmManager.enabled = newEnabled;

		var editorBody = rsEditor.getBody();

		if (newEnabled) {
			// If we get undefind or garbage value for the existing top padding, just plan on resetting it back to empty
			var newOriginalTopPaddingValue = ""
			var topPaddingValue = editorBody.style.paddingTop;
			if (topPaddingValue != null) {
				newOriginalTopPaddingValue = topPaddingValue;
			}
			tmManager.originalTopPadding = newOriginalTopPaddingValue;

			var newOriginalBottomPaddingValue = ""
			var bottomPaddingValue = editorBody.style.paddingBottom;
			if (bottomPaddingValue != null) {
				newOriginalTopPaddingValue = bottomPaddingValue;
			}
			tmManager.originalBottomPadding = newOriginalBottomPaddingValue;

			editorBody.style.paddingTop = "50%";
			editorBody.style.paddingBottom = "50%";
		}
		else {
			// See comments in -[RSHTMLSourceTextView resetViewsForNonTypewriterMode]. Same goal applies here, but it's harder
			// to just account for the size of the padding, because we don't know how much it is exactly (it's just "50%"). We
			// can ask the DOM for the computed padding size though. The regex replace at the end strips off non-digit values
			// e.g. "px" that is likely to be there, so we can ask for the resulting string's floatValue.
			var computedStyle = rsEditor.getWin().getComputedStyle(editorBody, null);
			var topPaddingHeight = computedStyle.getPropertyValue('padding-top').replace(/[^-\\d\\.]/g, '');

			// Need to adapt RSTinyMCEDOMEditor's scroll adjustment?

			// Restore original padding
			editorBody.style.paddingTop = tmManager.originalTopPadding;
			editorBody.style.paddingBottom = tmManager.originalBottomPadding;
		}

		// Ensure the client knows about the adjusted position of the active selection rect
		rsSendSelectionChangedNotification();
	}
	return "";
}

// Typewriter mode support. This returns the rect containing whatever the current selection is.
function rsSelectionBoundingRect() {

	var boundingRect = {x: 0, y: 0, width: 0, height: 0};

	function boundingRectFromRectList(rectList) {
		var boundingRect = new DOMRect();
		var rectCount = 0;
		if (rectList && rectList.length) {
			rectCount = rectList.length;
		}

		function maxX(rect) {
			return rect.x + rect.width;
		}

		function maxY(rect) {
			return rect.y + rect.height;
		}

		// Cumulatively union the rects together
		for (i = 0; i < rectList.length; i++) {
			var thisRect = rectList[i];

			// Adjust thisRect to reflect the absolute position within the document. getClientRects
			// adjusts based on the scroll offset
			thisRect.x += rsEditor.getDoc().documentElement.scrollLeft;
			thisRect.y += rsEditor.getDoc().documentElement.scrollTop;

			if (i == 0) {
				boundingRect = thisRect;
			}
			else {
				var unionX = Math.min(boundingRect.x, thisRect.x);
				var unionY = Math.min(boundingRect.y, thisRect.y);
				var unionWidth = Math.max(boundingRect.right, thisRect.right) - unionX;
				var unionHeight = Math.max(maxY(boundingRect), maxY(thisRect)) - unionY;
				boundingRect.x = unionX;
				boundingRect.y = unionY;
				boundingRect.width = unionWidth;
				boundingRect.height = unionHeight;
			}
		}
		return boundingRect;
	}

	// getBoundingClientRect would seem most suitable, but in my tests the method returns a rect
	// filled with zeroes when the selection is a collapsed insertion point. We have to do some smarter
	// stuff with getClientRects() instead.
	var selection = rsEditor.getWin().getSelection();
	if (selection.rangeCount > 0) {
		var rectList = selection.getRangeAt(0).getClientRects();

		// If we are an empty cursor in an empty paragraph for example the rectList will be empty
		if (rectList.length === 0) {
			var el = rsSelectedElement();
			if (el) {
				boundingRect = boundingRectFromRectList(rsSelectedElement().getClientRects());
			}
		}
		else {
			boundingRect = boundingRectFromRectList(rectList);
		}
	}

	return boundingRect;
}
